# -*- coding: utf-8 -*-
"""
Intro
=====
Experimental

Sanskrit pada generator

Process
1. PrakRiti (dhatu / prAtipadika)
2. pratyaya addition
  2.a strIpratyaya
  2.b samAsa
3. prakRti / pratyaya transformations
4. aNga transformations
5. prakRti + pratyaya sandhi
6. pada sandhi

@author: Karthik Madathil (github: @kmadathil)
"""
